$LOAD_PATH.unshift __dir__

require "runc/version"
require 'net/http'
require 'readline'
require 'yaml'

##
# nc REPL
class Runc
  # :stopdoc:

  attr_reader :header, :port

  attr_accessor :protocol, :req, :uri

  %w(INT TERM).each do |sig|
    trap(sig) {puts; exit}
  end

  def self.main(args)
    raise(RuntimeError, "Missing nc dependency") unless has_nc

    help if(args.include?("-h") or args.include?("--help"))
    version if (args.include?("-v") or args.include?("--version"))

    if(args.include?('-p') or args.include?('--port'))
      args.map! {|a| a == '--interval' ? '-i' : a}
      _, port = args.slice!(args.index('-p'),2)
      port = port.to_i
    end
    port ||= 8000

    if(args.include?('-n') or args.include?('--host'))
      args.map! {|a| a == '--host' ? '-n' : a}
      _, hostname = args.slice!(args.index('-n'),2)
    end
    hostname ||= 'localhost'

    body = (args.include?("-B") or args.include?("--no-body")) ? nil : true
    res_header = (args.include?("-H") or args.include?("--no-header")) ? nil : true

    new(hostname, port, body, res_header).repl
  rescue => e
    abort e.message
  end

  # :startdoc:

  ##
  # check if nc is installed
  def self.has_nc
    system("type -p nc", [:out, :err] => File::NULL)
  end

  def self.help # :nodoc:
    puts <<~eos
    Usage:
      runc [options]

    Example:
      runc
      header> host: example.com, port: 80
      request> get

    Options:
      -B, --no-body     discard the response body
      -H, --no-header   discard the response header
      -h, --help        print this message
      -n, --host        host name (default 'localhost')
      -p, --port        port (default 8000)
      -v, --version     print runc version
    eos
    exit
  end

  def self.version # :nodoc:
    puts "runc #{VERSION}"
    exit
  end

  def initialize(hostname, port, body, res_header) # :nodoc:
    @header = {'host' => hostname}
    @port = port
    @body = body
    @res_header = res_header
    @req = 'get'
  end

  ##
  # file with internet protocols and ports. More info in 'man services'
  def protocol_file
    '/etc/services'
  end

  ##
  # generate a hash of protocols and port numbers
  def read_protocols
    @protocol = {}
    File.open(protocol_file) do |f|
      f.each_line do |l|
        next if l.match?(/^#|$^/)
        @protocol[l.split[1].sub(/\/.*/i,'').to_i] = l.split[0]
      end
    end

    # default to http if +port+ is larger than 2000
    @protocol.keys.each do |port|
      @protocol[port] = 'http' if port > 2000
    end
  end

  def repl # :nodoc:
    read_protocols
    loop do
      update_header

      req_tmp=ask('request')

      @req = req_tmp.empty? ? @req : req_tmp

      update_uri

      request=Net::HTTP.const_get(@req[/^\w+/].capitalize).new(@uri, @header)

      if [Net::HTTP::Post, Net::HTTP::Put, Net::HTTP::Patch].include?(request.class)
        request.body=ask('body')
      end

      http=Net::HTTP.new(@uri.host,@uri.port)

      http.use_ssl = port==443 ? true : false

      response=http.request(request)

      # request headers
      @header.map {|h, v| puts "> \e[1m#{h}\e[m: #{v}"}
      puts ">"

      # response headers
      if @res_header
        puts "< HTTP/1.1 #{response.code} #{response.message}"
        response.each_header {|h, v| puts "< \e[1m#{h}\e[m: #{v}"}
        puts "<"
      end

      # response body
      puts(response.body) if @body
    rescue => e
      puts "#{caller.first}: #{e.message}"
    end
  end

  ##
  # reads a hash from user input
  # hash should be of the form: key1: val1 [, key2: val2 [, ...]]
  def update_header
    header_tmp=YAML.load("{#{ask('header')}}")

    # if +header_tmp+ contains a +port+ key, update the port
    @port=header_tmp.key?('port') ? header_tmp.delete('port') : @port

    @header.update(header_tmp)
  end

  ##
  # prompt user for input
  def ask(prompt)
    res = Readline.readline("#{prompt}> ", true)
    res.nil? ? (puts; exit) : res
  end

  def update_uri # :nodoc:
    @uri=URI("#{@protocol[port]}://#{@header['host']}:#{port}/#{@req.sub(/^\w+ */,'')}")
  end
end
